@@ -7,7 +7,7 @@ use Carp ();
 package Net::OpenID::Server;
 
 use vars qw($VERSION $HAS_CRYPT_DSA $HAS_CRYPT_OPENSSL);
-$VERSION = "0.04";
+$VERSION = "0.05";
 
 use fields (
             'last_errcode',   # last error code we got
@@ -37,16 +37,22 @@ use fields (
                                # "trust_root", "return_to", "post_grant", and "is_identity"
 
             'args',            # that magic anything reference (returns a subref)
+
+            'server_secret',    # subref returning secret given $time
+            'secret_gen_interval',
+            'secret_expire_age',
             );
 
 use URI;
 use MIME::Base64 ();
-use Digest::SHA1 ();
+use Digest::SHA1 qw(sha1 sha1_hex);
+use Crypt::DH 0.05;
+use Math::BigInt;
 
 BEGIN {
     unless ($HAS_CRYPT_OPENSSL = eval "use Crypt::OpenSSL::DSA 0.12; 1;") {
         unless ($HAS_CRYPT_DSA = eval "use Crypt::DSA (); use Convert::PEM 0.07; 1;") {
-            die "You need Crypt::OpenSSL::DSA version 0.12+ -or- (Crypt::DSA -and- Convert::PEM 0.07+)\n";
+            warn "Old protocol mode requires Crypt::OpenSSL::DSA version 0.12+ -or- (Crypt::DSA -and- Convert::PEM 0.07+)\n";
         }
     }
 }
@@ -61,20 +67,24 @@ sub new {
 
     $self->args(delete $opts{args});
 
+    $opts{'secret_gen_interval'} ||= 86400;
+    $opts{'secret_expire_age'}   ||= 86400 * 14;
+
     $self->$_(delete $opts{$_})
         foreach (qw(
                     private_key public_key
                     get_user is_identity is_trusted
-                    setup_url setup_map
+                    setup_url setup_map server_secret
+                    secret_gen_interval secret_expire_age
                     ));
 
     Carp::croak("Unknown options: " . join(", ", keys %opts)) if %opts;
     return $self;
 }
 
-sub get_user    { &_getsetcode; }
-sub is_identity { &_getsetcode; }
-sub is_trusted  { &_getsetcode; }
+sub get_user     { &_getsetcode; }
+sub is_identity  { &_getsetcode; }
+sub is_trusted   { &_getsetcode; }
 
 sub public_key  { &_getsetkey; }
 sub private_key { &_getsetkey; }
@@ -82,6 +92,38 @@ sub private_key { &_getsetkey; }
 sub setup_url   { &_getset; }
 sub setup_map   { &_getset; }
 
+sub server_secret       { &_getset; }
+sub secret_gen_interval { &_getset; }
+sub secret_expire_age   { &_getset; }
+
+
+# returns ($content_type, $page), where $content_type can be "redirect"
+# in which case a temporary redirect should be done to the URL in $page
+# $content_type can also be "setup", in which case the setup_map variables
+# are in $page as a hashref, and caller has full control from there.
+#
+# returns undef on error, in which case caller should generate an error
+# page using info in $nos->err.
+sub handle_page {
+    my Net::OpenID::Server $self = shift;
+    my %opts = @_;
+    my $redirect_for_setup = delete $opts{'redirect_for_setup'};
+    Carp::croak("Unknown options: " . join(", ", keys %opts)) if %opts;
+    Carp::croak("handle_page must be called in list context") unless wantarray;
+
+    # OLDPROTO:
+    return $self->_mode_pubkey
+        if $self->args("openid.mode") eq "getpubkey";
+
+    return $self->_mode_associate
+        if $self->args("openid.mode") eq "associate";
+
+    my $mode = $self->args("openid.mode") || "checkid_immediate";
+    return $self->_fail("unknown_mode") unless $mode =~ /^checkid_(?:immediate|setup)/;
+
+    return $self->_mode_checkid($mode, $redirect_for_setup);
+}
+
 
 # given something that can have GET arguments, returns a subref to get them:
 #   Apache
@@ -122,6 +164,7 @@ sub args {
     $self->{args};
 }
 
+# OLDPROTO: will be removed
 sub signed_return {
     my Net::OpenID::Server $self = shift;
     my %opts = @_;
@@ -149,7 +192,6 @@ sub signed_return {
 
     my $sig64 = MIME::Base64::encode_base64($sig);
     chomp $sig64;  # remove \n
-
     my $ret_url = $return_to;
     _push_url_arg(\$ret_url,
                   "openid.mode",            "id_res",
@@ -160,26 +202,68 @@ sub signed_return {
     return $ret_url;
 }
 
-# returns ($content_type, $page), where $content_type can be "redirect"
-# in which case a temporary redirect should be done to the URL in $page
-# $content_type can also be "setup", in which case the setup_map variables
-# are in $page as a hashref, and caller has full control from there.
-#
-# returns undef on error, in which case caller should generate an error
-# page using info in $nos->err.
-sub handle_page {
+sub signed_return_url {
     my Net::OpenID::Server $self = shift;
     my %opts = @_;
-    my $redirect_for_setup = delete $opts{'redirect_for_setup'};
+    my $identity     = delete $opts{'identity'};
+    my $return_to    = delete $opts{'return_to'};
+    my $assoc_handle = delete $opts{'assoc_handle'};
+    my $assoc_type   = delete $opts{'assoc_type'} || "HMAC-SHA1";
+    my $trust_root   = delete $opts{'trust_root'};  # optional
     Carp::croak("Unknown options: " . join(", ", keys %opts)) if %opts;
-    Carp::croak("handle_page must be called in list context") unless wantarray;
 
-    return $self->_page_pubkey
-        if $self->args("openid.mode") eq "getpubkey";
+    if ($trust_root) {
+        return undef unless _url_is_under($trust_root, $return_to);
+    }
 
-    my $mode = $self->args("openid.mode") || "checkid_immediate";
+    # new protocol
+    my $ret_url = $return_to;
 
-    return $self->_fail("unknown_mode") unless $mode =~ /^checkid_(?:immediate|setup)/;
+    my $c_sec;
+    my $expires;
+    if ($assoc_handle) {
+        $c_sec = $self->_secret_of_handle($assoc_type, $assoc_handle)
+            or return undef;
+    } else {
+        # dumb consumer mode
+        ($assoc_handle, $c_sec, $expires) = $self->_generate_association($assoc_type);
+    }
+
+    my @sign = qw(mode issued valid_to identity return_to);
+    my $now = time();
+    my %arg = (
+               mode         => "id_res",
+               issued       => _time_to_w3c($now),
+               valid_to     => _time_to_w3c($now + 30),
+               identity     => $identity,
+               return_to    => $return_to,
+               assoc_type   => $assoc_type,
+               assoc_handle => $assoc_handle,
+               signed       => join(",", @sign),
+               );
+
+    my @arg; # arguments we'll append to the URL
+
+    my $token_contents = "";
+    foreach my $f (@sign) {
+        $token_contents .= "$f:$arg{$f}\n";
+        push @arg, "openid.$f" => $arg{$f};
+        delete $arg{$f};
+    }
+
+    # include the arguments we didn't sign in the URL
+    push @arg, map { ( "openid.$_" => $arg{$_} ) } sort keys %arg;
+
+    # finally include the signature
+    push @arg, "openid.sig" => _b64(hmac_sha1($token_contents, $c_sec));
+
+    _push_url_arg(\$ret_url, @arg);
+    return $ret_url;
+}
+
+sub _mode_checkid {
+    my Net::OpenID::Server $self = shift;
+    my ($mode, $redirect_for_setup) = @_;
 
     my $return_to = $self->args("openid.return_to");
     return $self->_fail("no_return_to") unless $return_to =~ m!^https?://!;
@@ -187,7 +271,17 @@ sub handle_page {
     my $trust_root = $self->args("openid.trust_root") || $return_to;
     return $self->_fail("invalid_trust_root") unless _url_is_under($trust_root, $return_to);
 
-    my $identity = $self->args("openid.is_identity");
+    my $identity = $self->args("openid.identity");
+
+    # OLDPROTO:
+    my $old_proto = 0;
+    unless ($identity) {
+        if ($identity = $self->args("openid.is_identity")) {
+            $old_proto = 1;
+        } else {
+            return $self->_fail("no_identity");
+        }
+    }
 
     # chop off the query string, in case our trust_root came from the return_to URL
     $trust_root =~ s/\?.*//;
@@ -196,13 +290,22 @@ sub handle_page {
     my $is_identity = $self->_proxy("is_identity", $u, $identity);
     my $is_trusted  = $self->_proxy("is_trusted",  $u, $trust_root, $is_identity);
 
-
     # assertion path:
     if ($is_identity && $is_trusted) {
-        my $ret_url = $self->signed_return(
-                                           identity => $identity,
-                                           return_to => $return_to,
-                                           );
+        my $ret_url;
+        if ($old_proto) {
+            $ret_url = $self->signed_return(
+                                               identity => $identity,
+                                               return_to => $return_to,
+                                               );
+        } else {
+            $ret_url = $self->signed_return_url(
+                                                identity => $identity,
+                                                return_to => $return_to,
+                                                assoc_handle => $self->args("openid.assoc_handle"),
+                                                assoc_type   => $self->args("openid.assoc_type"),
+                                                );
+        }
         return ("redirect", $ret_url);
     }
 
@@ -213,7 +316,7 @@ sub handle_page {
     my %setup_args = (
                       $self->_setup_map("trust_root"),  $trust_root,
                       $self->_setup_map("return_to"),   $return_to,
-                      $self->_setup_map("is_identity"), $identity,
+                      $self->_setup_map("identity"),    $identity,
                       );
 
     my $setup_url = $self->{setup_url} or Carp::croak("No setup_url defined.");
@@ -256,7 +359,64 @@ sub _proxy {
     return $getter->(@_);
 }
 
-sub _page_pubkey {
+sub get_server_secret {
+    my Net::OpenID::Server $self = shift;
+    my $time = shift;
+
+    my $ss;
+    if (ref $self->{server_secret} eq "CODE") {
+        $ss = $self->{server_secret};
+    } elsif ($self->{server_secret}) {
+        $ss = sub { return $self->{server_secret}; };
+    } else {
+        Carp::croak("You haven't defined a server_secret value or subref defined.\n");
+    }
+
+    my $sec = $ss->($time);
+    Carp::croak("Server secret too long") if length($sec) > 255;
+    return $sec;
+}
+
+# returns ($assoc_handle, $secret, $expires)
+sub _generate_association {
+    my Net::OpenID::Server $self = shift;
+    my $type = shift;
+    die unless $type eq "HMAC-SHA1";
+
+    my $now = time();
+    my $sec_time = $now - ($now % $self->secret_gen_interval);
+
+    my $s_sec = $self->get_server_secret($sec_time)
+        or Carp::croak("server_secret didn't return a secret given what should've been a valid time ($sec_time)\n");
+
+    my $nonce = _rand_chars(20);
+    my $handle = "$now:$nonce";
+    $handle .= ":" . substr(hmac_sha1_hex($handle, $s_sec), 0, 10);
+
+    my $c_sec = $self->_secret_of_handle($type, $handle, 0)
+        or return ();
+
+    my $expires = $sec_time + $self->secret_expire_age;
+    return ($handle, $c_sec, $expires);
+}
+
+sub _secret_of_handle {
+    my Net::OpenID::Server $self = shift;
+    my ($type, $handle, $no_verify) = @_;
+    die "type unsupported" unless $type eq "HMAC-SHA1";
+
+    my ($time, $nonce, $nonce_sig80) = split(/:/, $handle);
+
+    my $sec_time = $time - ($time % $self->secret_gen_interval);
+    my $s_sec = $self->get_server_secret($sec_time)  or return;
+    length($nonce)       == 20 or return;
+    length($nonce_sig80) == 10 or return;
+    return unless $no_verify || $nonce_sig80 eq substr(hmac_sha1_hex("$time:$nonce", $s_sec), 0, 10);
+
+    return hmac_sha1($handle, $s_sec);
+}
+
+sub _mode_pubkey {
     my Net::OpenID::Server $self = shift;
     my $pubkey = $self->_get_key_contents("public")
         or return;
@@ -264,6 +424,78 @@ sub _page_pubkey {
     return ("text/plain", $pubkey);
 }
 
+sub _mode_associate {
+    my Net::OpenID::Server $self = shift;
+
+    my $now = time();
+    my %prop;
+
+    my $assoc_type = "HMAC-SHA1";
+    # FUTURE: protocol will let people choose their preferred authn scheme,
+    # in which case we see if we support any of them, and override the
+    # default value of HMAC-SHA1
+
+    my ($assoc_handle, $secret, $expires) =
+        $self->_generate_association($assoc_type);
+
+    # make expires absolute (it can be relative)
+    $expires += $now unless $expires > 1000000000;
+
+    $prop{'assoc_type'}   = $assoc_type;
+    $prop{'assoc_handle'} = $assoc_handle;
+    $prop{'expiry'}       = _time_to_w3c($expires);
+    $prop{'issued'}       = _time_to_w3c($now);
+
+    if ($self->args("openid.enc") eq "DH-SHA1") {
+
+        my $dh   = Crypt::DH->new;
+        my $p    = _arg2bi($self->args("openid.dh_modulus"));
+        my $g    = _arg2bi($self->args("openid.dh_gen"));
+        my $cpub = _arg2bi($self->args("openid.dh_gx"));
+
+        return $self->_error_page("invalid dh params")
+            unless $p > 10 && $g > 1 && $cpub;
+
+        $dh->p($p);
+        $dh->g($g);
+        $dh->generate_keys;
+
+        my $dh_sec = $dh->compute_secret($cpub);
+
+        $prop{'dh_gy'}      = _bi2arg($dh->pub_key);
+        $prop{'enc'}        = "DH-SHA1";
+        $prop{'enc_secret'} = _b64(sha1($secret) ^ sha1(_bi2bytes($dh_sec)));
+
+    } else {
+        $prop{'secret'} = _b64($secret);
+    }
+
+    return $self->_serialized_props(\%prop);
+}
+
+sub _b64 {
+    my $val = MIME::Base64::encode_base64($_[0]);
+    chomp $val;
+    return $val;
+}
+
+sub _error_page {
+    my Net::OpenID::Server $self = shift;
+    return $self->_serialized_props({ 'error_text' => $_[0] });
+}
+
+sub _serialized_props {
+    my Net::OpenID::Server $self = shift;
+    my $props = shift;
+
+    my $body = "";
+    foreach (sort keys %$props) {
+        $body .= "$_:$props->{$_}\n";
+    }
+
+    return ("text/plain", $body);
+}
+
 sub _get_key_contents {
     my Net::OpenID::Server $self = shift;
     my $key = shift;
@@ -353,7 +585,7 @@ sub errtext {
 
 sub _dsa_sig {
     my ($self, $plain) = @_;
-    my $msg = Digest::SHA1::sha1($plain);
+    my $msg = sha1($plain);
 
     if ($HAS_CRYPT_OPENSSL) {
         my $priv = $self->_get_key_contents("private")
@@ -462,7 +694,8 @@ sub _url_is_under {
 }
 
 sub _time_to_w3c {
-    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time());
+    my $time = shift || time();
+    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($time);
     $mon++;
     $year += 1900;
 
@@ -484,6 +717,65 @@ sub _push_url_arg {
     }
 }
 
+sub _bi2bytes {
+    my $bigint = shift;
+    die "Can't deal with negative numbers" if $bigint->is_negative;
+
+    my $bits = $bigint->as_bin;
+    die unless $bits =~ s/^0b//;
+
+    # prepend zeros to round to byte boundary, or to unset high bit
+    my $prepend = (8 - length($bits) % 8) || ($bits =~ /^1/ ? 8 : 0);
+    $bits = ("0" x $prepend) . $bits if $prepend;
+
+    return pack("B*", $bits);
+}
+
+sub _bi2arg {
+    my $b64 = MIME::Base64::encode_base64(_bi2bytes($_[0]));
+    chomp $b64;
+    return $b64;
+}
+
+sub _bytes2bi {
+    return Math::BigInt->new("0b" . unpack("B*", $_[0]));
+}
+
+sub _arg2bi {
+    # don't acccept base-64 encoded numbers over 700 bytes.  which means
+    # those over 4200 bits.
+    return Math::BigInt->new("0") if length($_[0]) > 700;
+    return _bytes2bi(MIME::Base64::decode_base64($_[0]));
+}
+
+sub _rand_chars
+{
+    my $length = shift;
+    my $chal = "";
+    my $digits = "abcdefghijklmnopqrstuvwzyzABCDEFGHIJKLMNOPQRSTUVWZYZ0123456789";
+    for (1..$length) {
+        $chal .= substr($digits, int(rand(62)), 1);
+    }
+    return $chal;
+}
+
+# From Digest::HMAC
+sub hmac_sha1_hex {
+    unpack("H*", &hmac_sha1);
+}
+sub hmac_sha1 {
+    hmac($_[0], $_[1], \&sha1, 64);
+}
+sub hmac {
+    my($data, $key, $hash_func, $block_size) = @_;
+    $block_size ||= 64;
+    $key = &$hash_func($key) if length($key) > $block_size;
+
+    my $k_ipad = $key ^ (chr(0x36) x $block_size);
+    my $k_opad = $key ^ (chr(0x5c) x $block_size);
+
+    &$hash_func($k_opad, &$hash_func($k_ipad, $data));
+}
 
 __END__
 
@@ -585,7 +877,7 @@ with the arguments from setup_map already appended.
 
 =back
 
-=item $url = $nos->B<signed_return>( %opts )
+=item $url = $nos->B<signed_return_url>( %opts )
 
 Generates a positive identity assertion URL that you'd redirect a user
 to.  Typically this would be after they've completed your setup_url.
@@ -689,6 +981,29 @@ either $u is undef, or $is_identity is false (this is the result of
 your previous is_identity callback), you should return 0.  But your
 callback is always run so you can avoid timing attacks, if you care.
 
+=item $nos->B<server_secret>($scalar)
+
+=item $nos->B<server_secret>($code)
+
+=item $code = $nos->B<server_secret>; ($secret) = $code->($time);
+
+The server secret is used to generate and sign lots of per-consumer
+secrets, and is never handed out directly.
+
+In the simplest (and least secure) form, you configure a static secret
+value with a scalar.  If you use this method and change the scalar
+value, all consumers that have cached their per-consumer secrets will
+start failing, since their secrets no longer work.
+
+The recommended usage, however, is to supply a subref that returns a
+secret based on the provided I<$time>, a unix timestamp.  And if one
+doesn't exist for that time, create, store and return it (with
+appropriate locking so you never return different secrets for the same
+time.)  Your secret can just be random characters, but it's your
+responsibility to do the locking and storage.
+
+Your secret may not exceed 255 characters.
+
 =item $nos->B<setup_url>($url)
 
 =item $url = $nos->B<setup_url>
@@ -706,7 +1021,7 @@ When this module gives a consumer site a user_setup_url from your
 provided setup_url, it also has to append a number of get parameters
 onto your setup_url, so your app based at that setup_url knows what it
 has to setup.  Those keys are named, by default, "trust_root",
-"return_to", "post_grant", and "is_identity".  If you don't like those
+"return_to", "post_grant", and "identity".  If you don't like those
 parameter names, this $hashref setup_map lets you change one or more
 of them.  The hashref's keys should be the default values, with values
 being the parameter names you want.
